# ------------------------------------------------------------
# NQPV_la.py
#
# linear algebra tools needed in this verifier
# ------------------------------------------------------------

from .tools import err

import numpy as np

EPS = 1e-7

error_info = ""
silent = False

def complex_norm(c):
    return np.sqrt(c * np.conj(c)).real

def check_unity(m, m_id):
    '''
    check whether tensor m is unitary
    m: tensor of shape (2,2,...,2), with the row indices in front of the column indices
    '''
    global error_info, silent

    if len(m.shape) % 2 == 1:
        error_info += err("Error: The dimension of '" + m_id + "' is invalid for an unitary.\n\n", silent)
        return False

    for dim in m.shape:
        if dim != 2:
            error_info += err("Error: The dimension of '" + m_id + "' is invalid for an unitary.\n\n", silent)
            return False
    
    # calculate the dim for matrix
    dim_m = 2**(len(m.shape)//2)
    matrix = m.reshape((dim_m, dim_m))

    # check the maximum of U^dagger @ U - I
    zero_check = (matrix @ np.transpose(np.conj(matrix))) - np.eye(dim_m)
    diff = np.sqrt(complex_norm(np.max(zero_check)))
    if diff > EPS:
        error_info += err("Error: '" + m_id + "' is not an unitary matrix.\n\n", silent)
        return False
    return True

def check_hermitian_predicate(m, m_id):
    '''
    check whether tensor m is hermitian and 0 <= m <= I
    m: tensor of shape (2,2,...,2), with the row indices in front of the column indices
    '''
    global error_info, silent
    if len(m.shape) % 2 == 1:
        error_info += err("Error: The dimension of '" + m_id + "' is invalid for an Hermitian operator.\n\n", silent)
        return False

    for dim in m.shape:
        if dim != 2:
            error_info += err("Error: The dimension of '" + m_id + "' is invalid for an Hermitian operator.\n\n", silent)
            return False
    
    # calculate the dim for matrix
    dim_m = 2**(len(m.shape)//2)
    matrix = m.reshape((dim_m, dim_m))

    # check the maximum of U^dagger @ U - I
    zero_check = matrix - np.transpose(np.conj(matrix))
    diff = np.sqrt(complex_norm(np.max(zero_check)))
    if diff > EPS:
        error_info += err("Error: '" + m_id + "' is not an Hermitian operator.\n\n", silent)
        return False

    # check 0 <= matrix <= I
    e_vals = np.linalg.eigvals(matrix)
    if np.any(e_vals < 0 - EPS) or np.any(e_vals > 1 + EPS):
        error_info += err("Error: The requirement 0 <= '" + m_id + "' <= I is not satisfied.\n\n", silent)
        return False
        
    return True


def check_measure(m, m_id):
    '''
    check whether tensor m is a valid measurement
    m: tensor of shape (2,2,...,2), with the row indices in front of the column indices. 
        The first index of m corresponds to measurement result 0 or 1.
    '''
    global error_info, silent
    if len(m.shape) % 2 == 0:
        error_info += err("Error: The dimension of '" + m_id + "' is invalid for a measurement.\n\n", silent)
        return False

    for dim in m.shape:
        if dim != 2:
            error_info += err("Error: The dimension of '" + m_id + "' is invalid for a measurement.\n\n", silent)
            return False
    
    # calculate the dim for matrix
    dim_m = 2**((len(m.shape)-1)//2)

    # pick out M0 and M1
    m0 = m[0].reshape((dim_m, dim_m))
    m1 = m[1].reshape((dim_m, dim_m))

    # check the maximum of U^dagger @ U - I
    zero_check = (m0.conj().transpose() @ m0 + m1.conj().transpose() @ m1) - np.eye(dim_m)
    diff = np.sqrt(complex_norm(np.max(zero_check)))
    if diff > EPS:
        error_info += err("Error: '" + m_id + "' does not satisfy the normalization requirement of a measurement.\n\n", silent)
        return False
    return True
    
def eye_tensor(qubitn):
    '''
    return the identity matrix of 'qubitn' qubits, in the form of a (2,2,2,...) tensor, row indices in the front.
    '''
    return np.eye(2**qubitn).reshape((2,)*qubitn*2)

def hermitian_contract(qvar: list, H, qvar_act : list, M):
    '''
    conduct the transformation M.H.M^dagger and return the result hermitian operator
    qvar: name sequence of qubits in H
    qvar_act: name sequence of qubits in M

    Note: the operators H and M should have been checked already

    [index sequence of tensor H (and M)]

            qvar == [q1, q2, q3, ... , qn]

              n  n+1 n+2 n+3     2n-2 2n-1
              |   |   |   |  ...  |   |
             ---------------------------
            | q1  q2  q3     ...      qn|
             ---------------------------
              |   |   |   |  ...  |   |
              0   1   2   3      n-2 n-1

    '''
    nH = len(qvar)
    nM = len(qvar_act)
    # decide the indices for contraction. note that M^dagger is accessed by its conjugate and the same index list iM_ls
    iM_ls = list(range(nM, 2*nM))
    iH_left_ls = [qvar.index(v) for v in qvar_act]
    iH_right_ls = [i + nH for i in iH_left_ls]

    # decide the rearrangements, since the standard rearrangement is not what we want
    count_rem_MH = 0
    count_rem_HMd = nH
    rearrange_MH = []
    rearrange_HMd = []
    for i in range(nH):
        if i in iH_left_ls:
            rearrange_MH.append(2*nH-nM + qvar_act.index(qvar[i]))
        else:
            rearrange_MH.append(count_rem_MH)
            count_rem_MH += 1
        if i + nH in iH_right_ls:
            rearrange_HMd.append(2*nH-nM + qvar_act.index(qvar[i - nH]))
        else:
            rearrange_HMd.append(count_rem_HMd)
            count_rem_HMd += 1

    rearrange_MH = rearrange_MH + list(range(nH - nM, 2*nH - nM))
    rearrange_HMd = list(range(nH)) + rearrange_HMd

    # conduct the contraction and rearrange the indices
    temp1 = np.tensordot(H, M, (iH_left_ls, iM_ls)).transpose(rearrange_MH)
    temp2 = np.tensordot(temp1, np.conjugate(M), (iH_right_ls, iM_ls)).transpose(rearrange_HMd)
    return temp2

def dagger(M):
    '''
    for a tensor M with shape (2,2,2,...), return M^dagger
    Note: M should have been already checked
    '''
    nM = len(M.shape)//2
    trans = list(range(nM, nM*2)) + list(range(0, nM))
    return np.conjugate(M).transpose(trans)
    

def hermitian_init(qvar: list, H, qvar_init: list):
    '''
    initialize hermitian operator H at variables 'qvar_init'
    '''
    P0 = np.array([[1., 0.],
                    [0., 0.]])
    P1 = np.array([[0., 0.],
                    [1., 0.]])
    # initialize all the variables in order
    tempH = H
    for var in qvar_init:
        a = hermitian_contract(qvar, tempH, [var], P0)
        b = hermitian_contract(qvar, tempH, [var], P1)
        tempH = a + b
    
    return tempH

def tensor_to_matrix(t):
    nM = len(t.shape)//2
    ndim = 2**nM
    return t.reshape((ndim, ndim))


def hermitian_extend(qvar: list, H, qvar_H: list):
    '''
    extend the given hermitian operator, according to all variables qvar, and return
    '''
    nAll = len(qvar)
    nH = len(qvar_H)
    m_I = eye_tensor(nAll - nH)

    temp = np.tensordot(H, m_I, ([],[]))

    # rearrange the indices
    count_ext = 0
    r_left = []
    r_right = []
    for i in range(nAll):
        if qvar[i] in qvar_H:
            pos = qvar_H.index(qvar[i])
            r_left.append(pos)
            r_right.append(nH + pos)
        else:
            r_left.append(2*nH + count_ext)
            r_right.append(nAll + nH + count_ext)
            count_ext += 1
    
    return temp.transpose(r_left + r_right)



# ------------------------------------------------------------
# backward_transformer.py
#
# backward_transformer for nondeterministic quantum programs
# it calculates the weakest (liberal) precondition of a given program abstract syntax tree
# ------------------------------------------------------------

from .tools import err

from .NQPV_ast import *
from . import NQPV_la
from .set_order import sqsubseteq

error_info = ""
silent = False

def wlp_verify(prog : dict, pinfo : dict, preserve_pre):
    '''
    calculates the weakest liberal precondition
    prog: the given abstract syntax tree
    pinfo: information about the program (operators and so on)
    preserve_pre: whether to preserve intermediate preconditions
    return: the weakest liberal precondition (a set of hermitian operators)
    '''
    qvar = prog['qvar']

    # prepare the hermitian operators
    input_post = [NQPV_la.hermitian_extend(qvar, pinfo['herm'][tuple[0]], tuple[1]) for tuple in prog['post-cond']]
    input_pre = [NQPV_la.hermitian_extend(qvar, pinfo['herm'][tuple[0]], tuple[1]) for tuple in prog['pre-cond']]

    # add the information for intermediate preconditions
    if preserve_pre:
        pinfo['im_pre-cond'] = {}
        pinfo['im_pre-cond-no'] = 0

    calculated_pre = wlp_iter(prog['sequence'], prog['qvar'], input_post, pinfo, preserve_pre)

    if calculated_pre is not None:
        if sqsubseteq(input_pre, calculated_pre):
            return True
    
    return False

def insert_im_precond(pinfo: dict, pre_m_ls : list):
    '''
    pre_m_ls: preconditions to be inserted
    return the newly created identifiers
    '''
    new_id_ls = []
    for pre in pre_m_ls:
        # create a valid id
        lb = 'PRE'+str(pinfo['im_pre-cond-no'])
        while lb in pinfo['unitary'] or \
            lb in pinfo['measure'] or \
            lb in pinfo['herm'] or\
            lb in pinfo['im_pre-cond']:

            pinfo['im_pre-cond-no'] += 1
            lb = 'PRE'+str(pinfo['im_pre-cond-no'])
        
        pinfo['im_pre-cond'][lb] = pre
        new_id_ls.append(lb)

    return new_id_ls
    


def wlp_iter(sequence: list, qvar: list, postcond: list, pinfo: dict, preserve_pre):
    '''
    input:
        sequence: sequence of program sentences
        qvar: list of all quantum variables
        postcond: the post condition, in the cylinder extension form
        preserve_pre: whether to preserve intermediate preconditions
    output:
        weakest liberal precondition, the result
    '''
    global error_info, silent

    cur_postcond = postcond
    # transform in the reverse order
    for i in range(len(sequence)-1, -1, -1):

        sentence = sequence[i]
        if sentence.label == 'SKIP':
            pass

        elif sentence.label == 'ABORT':
            cur_postcond = [NQPV_la.eye_tensor(len(qvar))]

        elif sentence.label == 'INIT':
            cur_postcond = [NQPV_la.hermitian_init(qvar, H, sentence.vls)
                            for H in cur_postcond]

        elif sentence.label == 'UNITARY':
            Ud = NQPV_la.dagger(pinfo['unitary'][sentence.unitary])
            cur_postcond = [NQPV_la.hermitian_contract(qvar, H, sentence.vls, Ud)
                            for H in cur_postcond]

        elif sentence.label == 'IF':
            temp = []
            for H in cur_postcond:
                # then - subclause
                pre1 = wlp_iter(sentence.S1, qvar, [H], pinfo, preserve_pre)
                if pre1 is None:
                    return None
                # else - subclause
                pre0 = wlp_iter(sentence.S0, qvar, [H], pinfo, preserve_pre)
                if pre0 is None:
                    return None

                for H0 in pre0:
                    for H1 in pre1:
                        temp.append(
                            NQPV_la.hermitian_contract(qvar, H0, sentence.measure_vls,
                             pinfo['measure'][sentence.measure][0]) + 
                            NQPV_la.hermitian_contract(qvar, H1, sentence.measure_vls,
                             pinfo['measure'][sentence.measure][1])
                        )

            cur_postcond = temp
        
        elif sentence.label == 'WHILE':
            # Note: in this step, the weakest liberal precondition is partially determined by the given invariant
            temp = []

            #prepare the hermitian operators
            inv = [NQPV_la.hermitian_extend(qvar, pinfo['herm'][tuple[0]], tuple[1]) for tuple in sentence.inv]

            for H in cur_postcond:
                proposed_pre = []
                # check whether it is a valid invariant
                for Hinv in inv:
                    proposed_pre.append(
                        NQPV_la.hermitian_contract(qvar, H, sentence.measure_vls,
                         pinfo['measure'][sentence.measure][0]) +
                        NQPV_la.hermitian_contract(qvar, Hinv, sentence.measure_vls,
                         pinfo['measure'][sentence.measure][1])
                    )
                calculated_pre = wlp_iter(sentence.S, qvar, proposed_pre, pinfo, preserve_pre)
                if calculated_pre is None:
                    return None

                if sqsubseteq(inv, calculated_pre):
                    temp += proposed_pre
                else:
                    error_info += err("(line "+str(sentence.lineno)+")\tFailure: The given invariant '" + predicate_to_code(sentence.inv) + "' is invalid.\n\n", silent)
                    return None

            cur_postcond = temp

        
        elif sentence.label == 'NONDET_CHOICE':
            pre0 = wlp_iter(sentence.S0, qvar, cur_postcond, pinfo, preserve_pre)
            if pre0 is None:
                return None
            pre1 = wlp_iter(sentence.S1, qvar, cur_postcond, pinfo, preserve_pre)
            if pre1 is None:
                return None
            cur_postcond = pre0 + pre1

        else:
            raise Exception("Unknown program structure.")
        
        if preserve_pre:
            sentence.pre_cond = insert_im_precond(pinfo, cur_postcond)


    return cur_postcond



# ------------------------------------------------------------
# semantics_analyser.py
#
# semantics analyser for nondeterministic quantum programs
#
# verification include: 
#   - whether the variables are predefined
#   - whether the unitary/measurement/hermitian operator referred in the program exists in the folder
#   - whether the unitary/measurement/hermitian operator provided is valid
#   - whether the unitary/measurement/hermitian operataor matches the variable lists
#
# ------------------------------------------------------------

import numpy as np
from . import NQPV_la

# store the error information
from .tools import err
error_info = ""
silent = False

#check whether the variables in check_ls are predefined in var_ls
def check_variable(var_ls : list, check_ls : list):
    global error_info, silent

    used_qvar = []
    for var in check_ls:
        # check whether the variable is defined in qvar
        if var not in var_ls:
            error_info += err("Error: variable '" + var + "' not defined in qvar\n", silent)
            return False
        # check whether the variable has been used
        if var in used_qvar:
            error_info += err("Error: Variable '" + var + "' appeared more than once in the variable list "+str(check_ls) + ".\n", silent)
            return False
        used_qvar.append(var)
    return True

def check_unitary(id, unitary_dict, var_ls, run_path, lib_path):
    global error_info, silent

    if id in unitary_dict:
        pass
    else:
        try:
            loaded = np.load(lib_path + "/unitary/" + id + ".npy")
        except:
            try:
                loaded = np.load(run_path + "/" + id + ".npy")
            except:
                error_info += err("Error: unitary '" + id +".npy' not found\n", silent)
                return False

        # check for valid unitary
        if not NQPV_la.check_unity(loaded, id):
            return False

        unitary_dict[id] = loaded

    # check the dimension informaion
    if (len(var_ls) * 2 == len(unitary_dict[id].shape)):
        return True
    else:
        error_info += err("Error: The dimensions of unitary '" + id + "' and qvars " + str(var_ls) + " do not match.\n", silent)
        return False



def check_measure(id, measure_dict, var_ls, run_path, lib_path):
    global error_info, silent

    if id in measure_dict:
        pass
    else:
        try:
            loaded = np.load(lib_path + "/measure/" + id + ".npy")
        except:
            try:
                loaded = np.load(run_path + "/" + id + ".npy")
            except:
                error_info += err("Error: measurement '" + id +".npy' not found\n", silent)
                return False

        # check for valid measurement
        if not NQPV_la.check_measure(loaded, id):
            return False
    
        measure_dict[id] = loaded

    # check the dimension information
    # + 1 is for the extra dimension of 0-result and 1-result
    if (len(var_ls) * 2 + 1 == len(measure_dict[id].shape)):
        return True
    else:
        error_info += err("Error: The dimensions of measurement '" + id + "' and qvars " + str(var_ls) + " do not match.\n", silent)
        return False


def check_hermitian(id, herm_dict, var_ls, run_path, lib_path):
    # check whether the hermitian is suitable
    global error_info, silent

    if id in herm_dict:
        pass
    else:
        try:
            loaded = np.load(lib_path + "/herm/" + id + ".npy")
        except:
            try:
                loaded = np.load(run_path + "/" + id + ".npy")
            except:
                error_info += err("Error: hermitian operator '" + id +".npy' not found\n", silent)
                return False

        # check for valid unitary
        if not NQPV_la.check_hermitian_predicate(loaded, id):
            return False

        herm_dict[id] = loaded

    # check the dimension informaion
    if (len(var_ls) * 2 == len(herm_dict[id].shape)):
        return True
    else:
        error_info += err("Error: The dimensions of hermitian '" + id + "' and qvars " + str(var_ls) + " do not match.\n", silent)
        return False


# check the semantics of prog. load and return the operators
def check(prog : dict, run_path, lib_path):
    global error_info, silent

    if prog == None:
        error_info += err("Error: The input abstract syntax tree is invalid.\n", silent)
        return None

    # check the declared qvar, ensure the uniqueness
    if not check_variable(prog['qvar'], prog['qvar']):
        return None

    pinfo = {
        'unitary' : {},
        'measure' : {},
        'herm' : {},
        'while_exists' : False
    }
    
    # check the pre-condition and post-condition
    for herm_tuple in prog['pre-cond']:
        if not check_variable(prog['qvar'], herm_tuple[1]):
            return None
        if not check_hermitian(herm_tuple[0], pinfo['herm'], herm_tuple[1], run_path, lib_path):
            return None
    for herm_tuple in prog['post-cond']:
        if not check_variable(prog['qvar'], herm_tuple[1]):
            return None
        if not check_hermitian(herm_tuple[0], pinfo['herm'], herm_tuple[1], run_path, lib_path):
            return None

    if check_iter(prog['qvar'], prog['sequence'], pinfo, run_path, lib_path) :
        return pinfo
    else:
        return None
    

def check_iter(var_ls : list, sequence : list, pinfo : dict, run_path, lib_path):
    for sentence in sequence:
        if sentence.label == 'SKIP':
            pass
        elif sentence.label == 'ABORT':
            pass
        elif sentence.label == 'INIT':
            # check the variables
            if not check_variable(var_ls, sentence.vls):
                return False
        elif sentence.label == 'UNITARY':
            # check the variables
            if not check_variable(var_ls, sentence.vls):
                return False
            # check the unitary
            if not check_unitary(sentence.unitary, pinfo['unitary'], sentence.vls, run_path, lib_path):
                return False

        elif sentence.label == 'IF':
            # check the variables
            if not check_variable(var_ls, sentence.measure_vls):
                return False
            # check the measure existence
            if not check_measure(sentence.measure, pinfo['measure'], sentence.measure_vls, run_path, lib_path):
                return False
            # check the two subprograms
            if not check_iter(var_ls, sentence.S0, pinfo, run_path, lib_path):
                return False
            if not check_iter(var_ls, sentence.S1, pinfo, run_path, lib_path):
                return False
        
        elif sentence.label == 'WHILE':
            pinfo['while_exists'] = True
            # check the hermitian operators
            for herm_tuple in sentence.inv:
                if not check_variable(var_ls, herm_tuple[1]):
                    return False
                if not check_hermitian(herm_tuple[0], pinfo['herm'], herm_tuple[1], run_path, lib_path):
                    return False
            # check the variables
            if not check_variable(var_ls, sentence.measure_vls):
                return False
            # check the measure existence
            if not check_measure(sentence.measure, pinfo['measure'], sentence.measure_vls, run_path, lib_path):
                return False
            # check the subprogram
            if not check_iter(var_ls, sentence.S, pinfo, run_path, lib_path):
                return False
        
        elif sentence.label == 'NONDET_CHOICE':
            # check the two subprograms
            if not check_iter(var_ls, sentence.S0, pinfo, run_path, lib_path):
                return False
            if not check_iter(var_ls, sentence.S1, pinfo, run_path, lib_path):
                return False
        
        else:
            raise Exception("Unknown program structure.")

    # after the examination of all programs in the sequence
    return True



# ------------------------------------------------------------
# set_order.py
#
# decide the partial order of Hermitian operator sets, using SDP
# ------------------------------------------------------------

from .NQPV_la import EPS

import numpy as np
import cvxpy as cp

def sqsubseteq(setA : list, setB : list) -> bool:
    '''
    decide the partial order of Hermitian operator sets, using SDP
    setA, setB: list of hermitian operators, in numpy array form, with shape of (2,2,2,2,...)
        These hermitian operators should have been examinated already.
    return: True if setA sqsubseteq setB, else False.
    '''
    
    # transform all the hermitian operators into matrices
    qubitn = len(setA[0].shape)//2
    dim = 2**qubitn
    msetA = [m.reshape((dim, dim)) for m in setA]
    msetB = [m.reshape((dim, dim)) for m in setB]

    #print(msetA)
    #print(msetB)

    for mB in msetB:
        X = cp.Variable((dim, dim), hermitian=True)
        constraints = [X >> 0]
        constraints += [
            cp.real(cp.trace((mB - mA) @ X)) <= -EPS for mA in msetA
        ]
        prob = cp.Problem(cp.Minimize(0),
                  constraints)
        prob.solve(eps = EPS/10)

        # Print result. debug purpose.
        '''
        print(constraints[-1])
        print("The optimal value is", prob.value)
        print("A solution X is")
        print(X.value)
        '''

        # if a solution has been found
        if X.value is not None:
            return False

    return True



# ------------------------------------------------------------
# NQPV_parser.py
#
# parser for nondeterministic quantum programs
# ------------------------------------------------------------

import ply.yacc as yacc

from .NQPV_lexer import tokens, lexer

from .NQPV_ast import *

# program declaration section
def p_prog(p):
    'prog : QVAR id_ls predicates_ls sequence predicates_ls'
    p[0] = {
        'qvar' : p[2],
        'pre-cond' : p[3],
        'sequence' : p[4],
        'post-cond' : p[5]
    }

# prog to code
def prog_to_code(p, prefix):
    r = prefix + "qvar "+ vls_to_code(p['qvar']) +"\n\n"
    r += prefix + "{ " + predicate_to_code(p['pre-cond']) + " }\n\n"
    r += sequence_to_code(p['sequence'], prefix) + "\n\n"
    r += prefix + "{ " + predicate_to_code(p['post-cond']) + " }\n"
    return r

def p_sequence_form(p):
    'sequence : sentence'
    p[0] = [p[1]]

def p_sequence_append(p):
    'sequence : sequence SEMICOLON sentence'
    p[0] = p[1] + [p[3]]


def p_sentence(p):
    '''
    sentence : skip
        | abort
        | init
        | unitary
        | if
        | while
        | nondet_choice
    '''
    p[0] = p[1]

def p_nondet_choice(p):
    'nondet_choice : LBRAKET sequence NONDET_CHOICE sequence RBRAKET'
    p[0] = NondetStruct(p)

def p_if(p):
    'if : IF ID id_ls THEN sequence ELSE sequence END'
    p[0] = IfStruct(p)

def p_while(p):
    'while : inv_ls WHILE ID id_ls DO sequence END'
    p[0] = WhileStruct(p)


def p_skip(p):
    'skip : SKIP'
    p[0] = SkipStruct(p)

def p_abort(p):
    'abort : ABORT'
    p[0] = AbortStruct(p)

def p_unitary(p):
    '''unitary : id_ls MUL_EQ ID
                | ID MUL_EQ ID'''
    p[0] = UnitaryStruct(p)

def p_init(p):
    '''init : id_ls ASSIGN ZERO
            | ID ASSIGN ZERO'''
    p[0] = InitStruct(p)

#define the invariants
def p_inv_end(p):
    'inv_ls : inv_pre RBRACE'
    p[0] = p[1]

def p_inv_append(p):
    'inv_pre : inv_pre ID id_ls'
    p[0] = p[1] + [(p[2], p[3])]

def p_inv_start(p):
    'inv_pre : LBRACE INV COLON ID id_ls'
    p[0] = [(p[4], p[5])]


#define the quantum predicates
def p_predicates_end(p):
    'predicates_ls : predicates_pre RBRACE'
    p[0] = p[1]

def p_predicates_append(p):
    'predicates_pre : predicates_pre ID id_ls'
    p[0] = p[1] + [(p[2], p[3])]

def p_predicates_start(p):
    'predicates_pre : LBRACE ID id_ls'
    p[0] = [(p[2], p[3])]

# define the list of quantum variables
def p_id_ls_end(p):
    'id_ls : id_ls_pre RSBRAKET'
    p[0] = p[1]

def p_id_ls_append(p):
    'id_ls_pre : id_ls_pre ID'
    p[0] = p[1] + [p[2]]

def p_id_ls_start(p):
    'id_ls_pre : LSBRAKET ID'
    p[0] = [p[2]]

# Error rule for syntax errors

from .tools import err

error_info = ""
silent = False

def p_error(p):
    global error_info, silent
    cur_info = "(line " + str(p.lineno) + ")\tSyntax error in input: '" + str(p.value) + "'. \n"
    error_info += err(cur_info, silent)


# Build the lexer
parser = yacc.yacc()


# ------------------------------------------------------------
# NQPV_verifier.py
#
# method 'verify' provides the user interface of the whole verifier
# ------------------------------------------------------------

import os
import numpy as np

from .tools import msg, ver_label, lineno_added

from .NQPV_ast import *
from . import NQPV_lexer, NQPV_parser, NQPV_la

from . import semantics_analyser
from .semantics_analyser import check
from . import backward_transformer
from .backward_transformer import wlp_verify

def verifiy_reset():
    # clean up the error info
    NQPV_la.error_info = ""
    NQPV_la.silent = True
    NQPV_lexer.error_info = ""
    NQPV_lexer.silent = True
    NQPV_parser.error_info = ""
    NQPV_parser.silent = True
    semantics_analyser.error_info = ""
    semantics_analyser.silent = True
    backward_transformer.error_info = ""
    backward_transformer.silent = True


def verify(folder_path, lib_path = "", silent = False, total_correctness = False, preserve_pre = False, opt_in_output = False, save_opt = False):

    if total_correctness:
        print("total correctness not supported yet")
        return

    verifiy_reset()

    msg("\n= Nondeterministic Quantum Program Verifier Output = \n\n", silent, None)

    msg("Version: " + ver_label + "\n\n", silent, None)

    # Prompt the running path
    msg("running path: " + os.getcwd() +"\n\n", silent, None)

    # detect the file
    if folder_path[-1] == '/' or folder_path[-1] == '\\':
        folder_path = folder_path[:-1]
    if lib_path != "":
        if lib_path[-1] == '/' or lib_path[-1] == '\\':
            lib_path += lib_path[:-1]

    
    try:
        p_prog = open(folder_path + '/prog', 'r')
        prog_str = p_prog.read()
        p_prog.close()
    except:
        print("Error: program file '" + folder_path + "/prog' not found.")
        return

    
    # create the output file
    try:
        p_output = open(folder_path + '/output.txt', 'w')
    except:
        print("Error: cannot create output file '" + folder_path + "/output.txt'.")
        return

    # add the beginning of output.txt
    p_output.write("\n= Nondeterministic Quantum Program Verifier Output = \n\n")
    p_output.write("Version: " + ver_label + "\n\n")
    p_output.write("running path: " + os.getcwd() +"\n\n")

    msg("folder path: "+ folder_path + "\n\n", silent, p_output)

    if total_correctness:
        msg("property to verify: Total Correctness\n\n", silent, p_output)
    else:
        msg("property to verify: Partial Correctness\n\n", silent, p_output)

    msg("intermediate preconditions preserved: "+("Yes" if preserve_pre else "No") + "\n\n", silent, p_output)
    msg("show operators in this output: " + ("Yes" if opt_in_output else "No") + "\n\n", silent, p_output)
    msg("operators saved in running paht: "+ ("Yes" if save_opt else "No") + "\n\n", silent, p_output)

    msg("--------------------------------------------\n", silent, p_output)
    msg("<prog> \n\n", silent, p_output)
    msg(lineno_added(prog_str), silent, p_output)
    msg("\n\n--------------------------------------------\n\n", silent, p_output)

    
    # syntactic analysis, produce the abstract syntax tree
    msg("syntactic analysis ...\n\n", silent, None)
    ast = NQPV_parser.parser.parse(prog_str)
    if NQPV_lexer.error_info != "" or NQPV_parser.error_info != "":
        if NQPV_lexer.error_info != "":
            msg(NQPV_lexer.error_info, False, p_output)
        else:
            msg(NQPV_parser.error_info, False, p_output)
        msg("\nAbort: syntactic analysis not passed.\n", False, p_output)
        p_output.close()
        return
    msg("syntactic analysis passed.\n\n", silent, p_output)

    # semantic analysis
    msg("semantic analysis ...\n\n", silent, None)
    pinfo = check(ast, folder_path, lib_path)
    if pinfo is None:
        # it means there is error in the semantic analysis
        if semantics_analyser.error_info != "":
            msg(semantics_analyser.error_info, False, p_output)
        elif NQPV_la.error_info != "":
            msg(NQPV_la.error_info, False, p_output)

        msg("\nAbort: semantic analysis not passed.\n", False, p_output)
        p_output.close()
        return
    msg("semantic analysis passed.\n\n", silent, p_output)


    # start verification

    msg("verification starts, calculating weakest (liberal) preconditions...\n\n", silent, None)

    v_result = wlp_verify(ast, pinfo, preserve_pre)

    if backward_transformer.error_info != "":
        msg(backward_transformer.error_info, False, p_output)

    # declare the result
    if v_result:
        msg("Verification Result: Property holds.\n\n", False, p_output)
    else:
        # check whether there is while structures
        if pinfo['while_exists']:
            msg("Verification Result: Property cannot be determined. A suitable loop invariant may be sufficient.\n\n", False, p_output)
        else:
            msg("Verification Result: Property does not hold.\n\n", False, p_output)

    # show the proof outline
    
    msg("(proof outline shown in 'output.txt)\n\n", silent, None)
    msg("--------------------------------------------\n", True, p_output)
    msg("<prog proof outline> \n\n", True, p_output)
    msg(lineno_added(NQPV_parser.prog_to_code(ast, "")), True, p_output)



    # show the operators in the output
    if opt_in_output:

        # prompt that the operator output is hidden in cmd
        if not silent:
            print("(operators shown in 'output.txt')\n\n", end='')
        
        msg("\n\n--------------------------------------------\n", True, p_output)
        msg("<unitary operators> \n\n", True, p_output)

        for id in pinfo['unitary']:
            msg(id + "\n", True, p_output)
            msg(str(NQPV_la.tensor_to_matrix(pinfo['unitary'][id])), True, p_output)
            msg("\n\n", True, p_output)
        
        msg("--------------------------------------------\n", True, p_output)

        msg("<Hermitian operators> \n\n", True, p_output)

        for id in pinfo['herm']:
            msg(id + "\n", True, p_output)
            msg(str(NQPV_la.tensor_to_matrix(pinfo['herm'][id])), True, p_output)
            msg("\n\n", True, p_output)
        
        msg("--------------------------------------------\n", True, p_output)

        msg("<measurement operators> \n\n", True, p_output)

        for id in pinfo['measure']:
            msg(id + " RESULT0 \n", True, p_output)
            msg(str(NQPV_la.tensor_to_matrix(pinfo['measure'][id][0])), True, p_output)
            msg("\n\n", True, p_output)
            msg(id + " RESULT1 \n", True, p_output)
            msg(str(NQPV_la.tensor_to_matrix(pinfo['measure'][id][1])), True, p_output)
            msg("\n\n", True, p_output)
        
        msg("--------------------------------------------\n", True, p_output)

        if preserve_pre:
            msg("<Hermitians for intermediate preconditions> \n\n", True, p_output)

            for id in pinfo['im_pre-cond']:
                msg(id + "\n", True, p_output)
                msg(str(NQPV_la.tensor_to_matrix(pinfo['im_pre-cond'][id])), True, p_output)
                msg("\n\n", True, p_output)
            
            msg("--------------------------------------------\n", True, p_output)

    # save the related operaters
    if save_opt:
        msg("Saving operators...\n\n", silent, None)

        for id in pinfo['unitary']:
            msg("unitary: " + id + ".npy ... ", silent, None)
            np.save(folder_path + "/" + id + ".npy", pinfo['unitary'][id])
            msg("done\n", silent, None)
        for id in pinfo['measure']:
            msg("measurement: " + id + ".npy ... ", silent, None)
            np.save(folder_path + "/" + id + ".npy", pinfo['measure'][id])
            msg("done\n", silent, None)
        for id in pinfo['herm']:
            msg("Hermitian Operator: " + id + ".npy ... ", silent, None)
            np.save(folder_path + "/" + id + ".npy", pinfo['herm'][id])
            msg("done\n", silent, None)
        if preserve_pre:
            for id in pinfo['im_pre-cond']:
                msg("Hermitian Operator: " + id + ".npy ... ", silent, None)
                np.save(folder_path + "/" + id + ".npy", pinfo['im_pre-cond'][id])
                msg("done\n", silent, None)


    # close the output file
    p_output.close()



# ------------------------------------------------------------
# NQPV_lexer.py
#
# tokenizer for nondeterministic quantum programs
# ------------------------------------------------------------
import ply.lex as lex

reserved = {
    'qvar' : 'QVAR',
    'skip': 'SKIP',
    'abort' : 'ABORT',
    'if': 'IF',
    'then': 'THEN',
    'else': 'ELSE',
    'while': 'WHILE',
    'do' : 'DO',
    'end' : 'END',
    'inv' : 'INV'
}

# List of token names.
tokens = [
    'ID',
    'ASSIGN',
    'ZERO',
    'MUL_EQ',
    'SEMICOLON',
    'NONDET_CHOICE',
    'COLON',
    'LSBRAKET',
    'RSBRAKET',
    'LBRAKET',
    'RBRAKET',
    'LBRACE',
    'RBRACE'
 ] + list(reserved.values())

# Regular expression rules for simple tokens
t_ASSIGN = r':='
t_ZERO = r'0'
t_MUL_EQ = r'\*='
t_SEMICOLON = r';'
t_NONDET_CHOICE = r'[#]'
t_COLON = r':'
t_LSBRAKET = r'\['
t_RSBRAKET = r'\]'
t_LBRAKET = r'\('
t_RBRAKET = r'\)'
t_LBRACE = r'{'
t_RBRACE = r'}'

# use // to comment a line
def t_COMMENT(t):
    r'//.*'
    pass

def t_ID(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'ID')    # Check for reserved words
    return t

# Define a rule so we can track line numbers
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)


# A string containing ignored characters (spaces and tabs)
t_ignore = ' \t'

# Error handling rule

from .tools import err

error_info = ""
silent = False

def t_error(t):
    global error_info, silent
    cur_info = "(line " + str(t.lineno) + ")\tSyntax Error. Illegal character '" + t.value[0] + "'. \n"
    error_info += err(cur_info, silent)
    t.lexer.skip(1)


# Build the lexer
lexer = lex.lex()

# test the lexer
if __name__ == '__main__':

    data = '''
    '''

    lexer.input(data)

    while True:
        tok = lexer.token()
        if not tok:
            break
        print(tok)


# ------------------------------------------------------------
# NQPV_ast.py
#
# definitions of the abstract syntax tree of NQPV
# ------------------------------------------------------------

def vls_to_code(vls):
    r = "[" + vls[0]
    for i in range(1, len(vls)):
        r += " " + vls[i]
    r += "]"
    return r

def predicate_to_code(predicate):
    r = predicate[0][0] + vls_to_code(predicate[0][1])
    for i in range(1, len(predicate)):
        r += " " + predicate[i][0] + vls_to_code(predicate[i][1])
    return r

def precondition_to_code(precond):
    '''
    precond: a list of identifiers(string), each represents a hermitian operator on the whole Hilbert space
    '''
    r = precond[0] + "[qvar]"
    for i in range(1, len(precond)):
        r += " " + precond[i] + "[qvar]"
    return r

class SyntaxStruct:
    # structure label
    label = ""
    # line number in the source code
    lineno = 0
    # precondition
    pre_cond = []

    def to_code(self, prefix):
        r = ""
        # if annoted, add the preconditions
        if self.pre_cond != []:
            r += "\n"+ prefix + "{ " + precondition_to_code(self.pre_cond) +" }\n"
        r += self.sentence_to_code(prefix)
        return r
    
    def sentence_to_code(self, prefix):
        return ""
        
class InitStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "INIT"
        self.lineno = p.slice[2].lineno
        if isinstance(p[1], list):
            self.vls = p[1]
        else:
            self.vls = [p[1]]

    def sentence_to_code(self, prefix):
        return prefix + vls_to_code(self.vls) + " := 0"

class UnitaryStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "UNITARY"
        self.lineno = p.slice[2].lineno
        self.unitary = p[3]
        if isinstance(p[1], list):
            self.vls = p[1]
        else:
            self.vls = [p[1]]

    def sentence_to_code(self, prefix):
        return prefix + vls_to_code(self.vls) + " *= " + self.unitary

class AbortStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "ABORT"
        self.lineno = p.slice[1].lineno

    def sentence_to_code(self, prefix):
        return prefix + "abort"


class SkipStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "SKIP"
        self.lineno = p.slice[1].lineno

    def sentence_to_code(self, prefix):
        return prefix + "skip"

class IfStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "IF"
        self.lineno = p.slice[1].lineno
        self.measure = p[2]
        self.measure_vls = p[3]
        self.S1 = p[5]  # then - subclause
        self.S0 = p[7]  # else - subclause
    
    def sentence_to_code(self, prefix):
        r = prefix + "if " + self.measure + vls_to_code(self.measure_vls) + " then\n"
        r += sequence_to_code(self.S1, prefix + '\t') + "\n"
        r += prefix + "else\n"
        r += sequence_to_code(self.S0, prefix + '\t') + "\n"
        r += prefix + "end"
        return r



class WhileStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "WHILE"
        self.lineno = p.slice[2].lineno
        self.inv = p[1]
        self.measure = p[3]
        self.measure_vls = p[4]
        self.S = p[6]
    
    def sentence_to_code(self, prefix):
        r = prefix + "{ inv: " + predicate_to_code(self.inv) + " }\n"
        r += prefix + "while " + self.measure + vls_to_code(self.measure_vls) + " do\n"
        r += sequence_to_code(self.S, prefix + '\t') + "\n"
        r += prefix + "end"
        return r

class NondetStruct(SyntaxStruct):
    def __init__(self, p):
        self.label = "NONDET_CHOICE"
        self.lineno = p.slice[3].lineno
        self.S0 = p[2]
        self.S1 = p[4]
    
    def sentence_to_code(self, prefix):
        r = prefix + "(\n"
        r += sequence_to_code(self.S0, prefix + '\t') + "\n"
        r += prefix + "#\n"
        r += sequence_to_code(self.S1, prefix + '\t') + "\n"
        r += prefix + ")"
        return r


def sequence_to_code(sequence: list, prefix):
    r = sequence[0].to_code(prefix)
    for i in range(1, len(sequence)):
        r += ";\n" + sequence[i].to_code(prefix)
    return r


# ------------------------------------------------------------
# NQPV_opt.py
#
# provides methods to create and verify 
# ------------------------------------------------------------

import numpy as np
from . import NQPV_la

# transform m to the (2,2,2,...) form
def to_shape_2(m):
    # check whether m already satisfies the requirements
    flag = True
    for dim in m.shape:
        if dim != 2:
            flag = False
            break
    if flag:
        return m
    
    if len(m.shape) == 2:
        if m.shape[0] != m.shape[1]:
            return None
        qubitn = int(np.log2(m.shape[0]))
        if m.shape[0] != 2**qubitn:
            return None
        return m.reshape((2,) * qubitn * 2)

    else:
        return None

# transform m to the (2,2,2,...) form (for measurements)
def to_shape_3(m):
    # check whether m already satisfies the requirements
    flag = True
    for dim in m.shape:
        if dim != 2:
            flag = False
            break
    if flag:
        return m
    
    if len(m.shape) == 3:
        if m.shape[1] != m.shape[2]:
            return None
        qubitn = int(np.log2(m.shape[1]))
        if m.shape[1] != 2**qubitn:
            return None
        return m.reshape((2,) * (qubitn * 2 + 1))

    else:
        return None


def save_unitary(path, id, unitary):
    m = to_shape_2(unitary)
    if m is None:
        print("The shape of the given tensor is invalid.")
        return False
    
    if not NQPV_la.check_unity(m, id):
        print("The given tensor is not unitary.")
        return False
    
    if path[-1] == '/' or path[-1] == '\\':
        path = path[:-1]
    
    np.save(path + "/" + id + ".npy", m)
    return True

def save_hermitian(path, id, herm):
    m = to_shape_2(herm)
    if m is None:
        print("The shape of the given tensor is invalid.")
        return False
    
    if not NQPV_la.check_hermitian_predicate(m, id):
        print("The given tensor is not a valid Hermitian predicate.")
        return False
    
    if path[-1] == '/' or path[-1] == '\\':
        path = path[:-1]
    
    np.save(path + "/" + id + ".npy", m)
    return True

def save_measurement(path, id, measure):
    m = to_shape_3(measure)
    if m is None:
        print("The shape of the given tensor is invalid.")
        return False
    
    if not NQPV_la.check_measure(m, id):
        print("The given tensor is not a valid measurement.")
        return False
    
    if path[-1] == '/' or path[-1] == '\\':
        path = path[:-1]
    
    np.save(path + "/" + id + ".npy", m)
    return True




# ------------------------------------------------------------
# tools.py
#
# define the tools used in this verifier
# ------------------------------------------------------------

ver_label = "0.1"

def err(txt, silent):
    if not silent:
        print(txt, end='')
    return txt


def msg(txt, silent, p_output):
    # message the text in the cmd and write in the output file
    if not silent:
        print(txt, end='')
    if p_output is not None:
        p_output.write(txt)

def lineno_added(txt):
    '''
    Add an line number in front of each line
    '''
    line_no = 1
    r = str(line_no) + '\t\t' + txt
    line_no += 1

    i = 0
    while i < len(r):
        if r[i] == '\n':
            r = r[:i+1] + str(line_no) + '\t\t' + r[i+1:]
            line_no += 1
        i += 1

    return r



from .NQPV_verifier import verify
from .opt_create import save_hermitian, save_measurement, save_unitary
from .lib_create import lib_create
from .NQPV_la import EPS
from .tools import ver_label as version